import pytest


def test_bad_subclass():
    from dataclasses import dataclass

    from jobflow import Maker

    @dataclass
    class BadMaker(Maker):
        name: str = "BadMaker"

    with pytest.raises(NotImplementedError):
        BadMaker().make()

    @dataclass
    class BadMaker(Maker):

        a = 1

    with pytest.raises(NotImplementedError):
        BadMaker().name()


def test_required_arguments_works():
    from dataclasses import dataclass

    @dataclass
    class MyMaker:

        a: int
        name = "123"

    m = MyMaker(1)
    assert m.a == 1

    with pytest.raises(TypeError):
        MyMaker()


def test_job_maker():
    from dataclasses import dataclass

    from jobflow.core.job import job
    from jobflow.core.maker import Maker

    @dataclass
    class AddMaker(Maker):
        name: str = "add"

        @job
        def make(self, a, b):
            return a + b

    maker = AddMaker()
    add_job = maker.make(1, 2)
    assert add_job.name == "add"
    assert add_job.function_args == (1, 2)
    assert add_job.maker == maker  # maker should match as all kwargs are the same
    assert str(add_job.function) == str(maker.make)

    # test updating maker class does not impact the job
    maker.name = "sum"
    assert add_job.maker != maker  # now makers should not match
    assert add_job.maker.name != "sum"


def test_flow_maker():
    from dataclasses import dataclass

    from jobflow import Flow, Maker, job

    @job
    def add(a, b):
        return a + b

    @dataclass
    class AddMaker(Maker):
        name: str = "add"

        def make(self, a, b):
            first = add(a, b)
            second = add(first.output, b)
            return Flow([first, second], second.output, name=self.name)

    maker = AddMaker()
    add_job = maker.make(1, 2)
    assert add_job.name == "add"
    assert len(add_job.jobs) == 2

    @dataclass
    class DoubleAddMaker(Maker):
        name: str = "add_add"
        add1: AddMaker = AddMaker()
        add2: AddMaker = AddMaker()

        def make(self, a, b):
            first = self.add1.make(a, b)
            second = self.add2.make(first.output, b)
            return Flow([first, second], second.output, name=self.name)

    maker = DoubleAddMaker()
    double_add_flow = maker.make(1, 2)
    assert double_add_flow.name == "add_add"
    assert len(double_add_flow.jobs) == 2
    assert isinstance(double_add_flow.jobs[0], Flow)
    assert isinstance(double_add_flow.jobs[1], Flow)


def test_update_kwargs():
    from dataclasses import dataclass

    from monty.json import MSONable

    from jobflow.core.job import Response, job
    from jobflow.core.maker import Maker

    # this is needed to get monty to deserialize them correctly
    global AddMaker
    global DetourMaker

    @dataclass
    class AddMaker(Maker):
        name: str = "add"
        c: int = 5

        @job
        def make(self, a, b):
            return a + b + self.c

    @dataclass
    class DetourMaker(Maker):
        name: str = "add"
        add_maker: Maker = AddMaker()

        def make(self, a, b):
            detour = self.add_maker.make(a, b)
            return Response(detour=detour)

    # test no filter
    maker = AddMaker()
    maker = maker.update_kwargs({"c": 10})
    assert maker.c == 10

    # test bad kwarg
    maker = AddMaker()
    with pytest.raises(TypeError):
        maker.update_kwargs({"d": 10})

    # test name filter
    maker = AddMaker()
    maker = maker.update_kwargs({"c": 10}, name_filter="add")
    assert maker.c == 10

    maker = AddMaker()
    maker = maker.update_kwargs({"c": 10}, name_filter="div")
    assert maker.c == 5

    # test class filter
    maker = AddMaker()
    maker = maker.update_kwargs({"c": 10}, class_filter=AddMaker)
    assert maker.c == 10

    maker = AddMaker()
    maker = maker.update_kwargs({"c": 10}, class_filter=maker)
    assert maker.c == 10

    maker = AddMaker()
    maker = maker.update_kwargs({"c": 10}, class_filter=list)
    assert maker.c == 5

    # test dict mod
    maker = AddMaker()
    maker = maker.update_kwargs({"_inc": {"c": 10}}, dict_mod=True)
    assert maker.c == 15

    # test nesting
    maker = DetourMaker()
    maker = maker.update_kwargs({"c": 10}, class_filter=AddMaker, nested=True)
    assert maker.add_maker.c == 10

    maker = DetourMaker()
    maker = maker.update_kwargs({"c": 10}, class_filter=AddMaker, nested=False)
    assert maker.add_maker.c == 5

    global NotAMaker
    global FakeDetourMaker

    @dataclass
    class NotAMaker(MSONable):
        name: str = "add"
        c: int = 5

        @job
        def make(self, a, b):
            return a + b + self.c

    @dataclass
    class FakeDetourMaker(Maker):
        name: str = "add"
        add_maker: MSONable = NotAMaker()

        def make(self, a, b):
            detour = self.add_maker.make(a, b)
            return Response(detour=detour)

    # test non maker dataclasses not updated
    maker = FakeDetourMaker()
    maker = maker.update_kwargs({"c": 10}, class_filter=NotAMaker, nested=True)
    assert maker.add_maker.c == 5



from dataclasses import dataclass

from jobflow import Flow, Maker, job, run_locally


@dataclass
class AddMaker(Maker):
    name: str = "add"
    c: int = 10

    @job
    def make(self, a, b, d=0.5):
        return a + b + self.c + d


maker = AddMaker(c=10)

add_first = maker.make(1, 2)
add_second = maker.make(add_first.output, 5)
flow = Flow([add_first, add_second])

flow.update_maker_kwargs({"_inc": {"c": 50}}, dict_mod=True)
flow.update_kwargs({"d": 0.2})

# run the flow, "output" contains the output of all jobs
output = run_locally(flow)
print(output)



"""
In this example, we show case the ability of jobflow to use multiple stores.

Jobs that use additional stores must be run with a JobStore that supports those stores.
"""

from maggma.stores import MemoryStore

from jobflow import JobStore, job, run_locally


@job(data=True)
def generate_big_data():
    """
    Generate some data.

    The data=True in the job decorator tells jobflow to store all outputs in the "data"
    additional store.
    """
    mydata = list(range(1000))
    return mydata


big_data_job = generate_big_data()

# in this example, we use different memory stores for the documents and "data"
# additional store. In practice, any Maggma Store subclass can be used for either store.
docs_store = MemoryStore()
data_store = MemoryStore()
store = JobStore(docs_store, additional_stores={"data": data_store})

# Because our job requires an additional store named "data" we have to use our
# custom store when running the job.
output = run_locally(big_data_job, store=store)

print(output)



"""A demonstration of how to enforce linear ordering of jobs."""
from jobflow import Flow, JobOrder, job


@job
def add(a, b):
    return a + b


# No edges between the jobs as they do not depend on each other
add_job_first = add(1, 2)
add_job_second = add(4, 6)
auto_flow = Flow([add_job_first, add_job_second], order=JobOrder.AUTO)
auto_flow.draw_graph().show()

# flow graph now shows an edge between the jobs due to the linear execution order
add_job_first = add(1, 2)
add_job_second = add(4, 6)
linear_flow = Flow([add_job_first, add_job_second], order=JobOrder.LINEAR)
linear_flow.draw_graph().show()



def test_find_key():
    from monty.json import MSONable, jsanitize

    from jobflow.utils import find_key

    data = {"a": [0, {"b": 1, "x": 3}], "c": {"d": {"x": 3}}}
    result = find_key(data, "x")
    assert len(result) == 2
    assert ["c", "d"] in result
    assert ["a", 1] in result

    result = find_key(data, "x", include_end=True)
    assert len(result) == 2
    assert ["c", "d", "x"] in result
    assert ["a", 1, "x"] in result

    data = {"a": {"x": {"x": 1}}, "b": {"x": 0}}
    result = find_key(data, "x", nested=False)
    assert len(result) == 2
    assert ["a"] in result
    assert ["b"] in result

    result = find_key(data, "x", nested=True)
    assert len(result) == 3
    assert ["a", "x"] in result
    assert ["a"] in result
    assert ["b"] in result

    class MyObj(MSONable):
        def __init__(self, a):
            self.a = a

    data = {"a": [0, {"b": 1, "x": 3}], "c": {"d": {"x": MyObj(a=1)}}}
    data = jsanitize(data, strict=True)
    assert find_key(data, MyObj) == [["c", "d", "x"]]


def test_find_key_value():
    from jobflow.utils import find_key_value

    data = {"a": [0, {"b": 1, "x": 3}], "c": {"d": {"x": 3}}}
    result = find_key_value(data, "x", 3)
    assert len(result) == 2
    assert ["a", 1] in result
    assert ["c", "d"] in result


def test_update_in_dictionary():
    from jobflow.utils import update_in_dictionary

    data = {"a": [0, {"b": 1, "x": 3}], "c": {"d": {"x": 3}}}
    update_in_dictionary(data, {("a", 1, "x"): 100, ("c", "d", "x"): 100})
    assert data == {"a": [0, {"b": 1, "x": 100}], "c": {"d": {"x": 100}}}


def test_contains_job_or_flow():
    from jobflow import Flow, Job
    from jobflow.utils import contains_flow_or_job

    job = Job(str)
    flow = Flow([])

    assert contains_flow_or_job(True) is False
    assert contains_flow_or_job(1) is False
    assert contains_flow_or_job("abc") is False
    assert contains_flow_or_job(job) is True
    assert contains_flow_or_job(flow) is True
    assert contains_flow_or_job([flow]) is True
    assert contains_flow_or_job([[flow]]) is True
    assert contains_flow_or_job({"a": flow}) is True
    assert contains_flow_or_job({"a": [flow]}) is True
    assert contains_flow_or_job(job) is True
    assert contains_flow_or_job([job]) is True
    assert contains_flow_or_job([[job]]) is True
    assert contains_flow_or_job({"a": job}) is True
    assert contains_flow_or_job({"a": [job]}) is True



import pytest


def test_simple_job(memory_jobstore, clean_dir, simple_job):
    from jobflow import run_locally

    # run with log
    job = simple_job("12345")
    uuid = job.uuid
    responses = run_locally(job, store=memory_jobstore)

    # check responses has been filled
    assert responses[uuid][1].output == "12345_end"

    # check store has the activity output
    result = memory_jobstore.query_one({"uuid": uuid})
    assert result["output"] == "12345_end"

    # test run no store
    job = simple_job("12345")
    uuid = job.uuid
    responses = run_locally(job)
    assert responses[uuid][1].output == "12345_end"


def test_simple_flow(memory_jobstore, clean_dir, simple_flow, capsys):
    from pathlib import Path

    from jobflow import run_locally

    flow = simple_flow()
    uuid = flow.jobs[0].uuid

    # run without log
    run_locally(flow, store=memory_jobstore, log=False)
    captured = capsys.readouterr()
    assert "INFO Started executing jobs locally" not in captured.out
    assert "INFO Finished executing jobs locally" not in captured.out

    # run with log
    responses = run_locally(flow, store=memory_jobstore)

    # check responses has been filled
    assert responses[uuid][1].output == "12345_end"

    # check store has the activity output
    result = memory_jobstore.query_one({"uuid": uuid})
    assert result["output"] == "12345_end"

    # check no folders were written
    folders = list(Path(".").glob("job_*/"))
    assert len(folders) == 0

    # check logs printed
    captured = capsys.readouterr()
    assert "INFO Started executing jobs locally" in captured.out
    assert "INFO Finished executing jobs locally" in captured.out

    # run with folders
    responses = run_locally(flow, store=memory_jobstore, create_folders=True)
    assert responses[uuid][1].output == "12345_end"
    folders = list(Path(".").glob("job_*/"))
    assert len(folders) == 1


def test_connected_flow(memory_jobstore, clean_dir, connected_flow):
    from jobflow import run_locally

    flow = connected_flow()
    uuid1 = flow.jobs[0].uuid
    uuid2 = flow.jobs[1].uuid

    # run with log
    responses = run_locally(flow, store=memory_jobstore)

    # check responses has been filled
    assert len(responses) == 2
    assert responses[uuid1][1].output == "12345_end"
    assert responses[uuid2][1].output == "12345_end_end"

    # check store has the activity output
    result1 = memory_jobstore.query_one({"uuid": uuid1})
    result2 = memory_jobstore.query_one({"uuid": uuid2})

    assert result1["output"] == "12345_end"
    assert result2["output"] == "12345_end_end"


def test_nested_flow(memory_jobstore, clean_dir, nested_flow):
    from jobflow import run_locally

    flow = nested_flow()
    uuid1 = flow.jobs[0].jobs[0].uuid
    uuid2 = flow.jobs[0].jobs[1].uuid
    uuid3 = flow.jobs[1].jobs[0].uuid
    uuid4 = flow.jobs[1].jobs[1].uuid

    # run with log
    responses = run_locally(flow, store=memory_jobstore)

    # check responses has been filled
    assert len(responses) == 4
    assert responses[uuid1][1].output == "12345_end"
    assert responses[uuid2][1].output == "12345_end_end"
    assert responses[uuid3][1].output == "12345_end_end_end"
    assert responses[uuid4][1].output == "12345_end_end_end_end"

    # check store has the activity output
    result1 = memory_jobstore.query_one({"uuid": uuid1})
    result2 = memory_jobstore.query_one({"uuid": uuid2})
    result3 = memory_jobstore.query_one({"uuid": uuid3})
    result4 = memory_jobstore.query_one({"uuid": uuid4})

    assert result1["output"] == "12345_end"
    assert result2["output"] == "12345_end_end"
    assert result3["output"] == "12345_end_end_end"
    assert result4["output"] == "12345_end_end_end_end"


def test_addition_flow(memory_jobstore, clean_dir, addition_flow):
    from jobflow import run_locally

    flow = addition_flow()
    uuid1 = flow.jobs[0].uuid

    # run with log
    responses = run_locally(flow, store=memory_jobstore)
    uuid2 = [u for u in responses.keys() if u != uuid1][0]

    # check responses has been filled
    assert len(responses) == 2
    assert responses[uuid1][1].output == 11
    assert responses[uuid1][1].addition is not None
    assert responses[uuid2][1].output == "11_end"

    # check store has the activity output
    result1 = memory_jobstore.query_one({"uuid": uuid1})
    result2 = memory_jobstore.query_one({"uuid": uuid2})

    assert result1["output"] == 11
    assert result2["output"] == "11_end"


def test_detour_flow(memory_jobstore, clean_dir, detour_flow):
    from jobflow import run_locally

    flow = detour_flow()
    uuid1 = flow.jobs[0].uuid
    uuid3 = flow.jobs[1].uuid

    # run with log
    responses = run_locally(flow, store=memory_jobstore)
    uuid2 = [u for u in responses.keys() if u != uuid1 and u != uuid3][0]

    # check responses has been filled
    assert len(responses) == 3
    assert responses[uuid1][1].output == 11
    assert responses[uuid1][1].detour is not None
    assert responses[uuid2][1].output == "11_end"
    assert responses[uuid3][1].output == "12345_end"

    # check store has the activity output
    result1 = memory_jobstore.query_one({"uuid": uuid1})
    result2 = memory_jobstore.query_one({"uuid": uuid2})
    result3 = memory_jobstore.query_one({"uuid": uuid3})

    assert result1["output"] == 11
    assert result2["output"] == "11_end"
    assert result3["output"] == "12345_end"

    # assert job2 (detoured job) ran before job3
    assert result2["completed_at"] < result3["completed_at"]


def test_replace_flow(memory_jobstore, clean_dir, replace_flow):
    from jobflow import run_locally

    flow = replace_flow()
    uuid1 = flow.jobs[0].uuid
    uuid2 = flow.jobs[1].uuid

    # run with log
    responses = run_locally(flow, store=memory_jobstore)

    # check responses has been filled
    assert len(responses) == 2
    assert len(responses[uuid1]) == 2
    assert responses[uuid1][1].output == 11
    assert responses[uuid1][1].replace is not None
    assert responses[uuid1][2].output == "11_end"
    assert responses[uuid2][1].output == "12345_end"

    # check store has the activity output
    result1 = memory_jobstore.query_one({"uuid": uuid1, "index": 1})
    result2 = memory_jobstore.query_one({"uuid": uuid1, "index": 2})
    result3 = memory_jobstore.query_one({"uuid": uuid2, "index": 1})

    assert result1["output"] == 11
    assert result2["output"] == "11_end"
    assert result3["output"] == "12345_end"

    # assert job2 (replaced job) ran before job3
    assert result2["completed_at"] < result3["completed_at"]


def test_replace_flow_nested(memory_jobstore, clean_dir, replace_flow_nested):
    from jobflow import run_locally

    flow = replace_flow_nested()
    uuid1 = flow.jobs[0].uuid
    uuid2 = flow.jobs[1].uuid

    # run with log
    responses = run_locally(flow, store=memory_jobstore)

    # check responses has been filled
    assert len(responses) == 4
    assert len(responses[uuid1]) == 2
    assert responses[uuid1][1].output == 11
    assert responses[uuid1][1].replace is not None
    assert responses[uuid1][2].output["first"].__class__.__name__ == "OutputReference"
    assert responses[uuid2][1].output == "12345_end"

    # check store has the activity output
    result1 = memory_jobstore.query_one({"uuid": uuid1, "index": 1})
    result2 = memory_jobstore.query_one({"uuid": uuid1, "index": 2})
    result3 = memory_jobstore.query_one({"uuid": uuid2, "index": 1})

    assert result1["output"] == 11
    assert result2["output"]["first"]["@class"] == "OutputReference"
    assert result3["output"] == "12345_end"

    # assert job2 (replaced job) ran before job3
    assert result2["completed_at"] < result3["completed_at"]


def test_stop_jobflow_flow(memory_jobstore, clean_dir, stop_jobflow_flow):
    from jobflow import run_locally

    flow = stop_jobflow_flow()
    uuid1 = flow.jobs[0].uuid

    # run with log
    responses = run_locally(flow, store=memory_jobstore)

    # check responses has been filled
    assert len(responses) == 1
    assert len(responses[uuid1]) == 1
    assert responses[uuid1][1].output == "1234"
    assert responses[uuid1][1].stop_jobflow is True

    # check store has the activity output
    result1 = memory_jobstore.query_one({"uuid": uuid1})

    assert result1["output"] == "1234"


def test_stop_jobflow_job(memory_jobstore, clean_dir, stop_jobflow_job):
    from jobflow import run_locally

    job = stop_jobflow_job()
    uuid1 = job.uuid

    # run with log
    responses = run_locally(job, store=memory_jobstore)

    # check responses has been filled
    assert len(responses) == 1
    assert len(responses[uuid1]) == 1
    assert responses[uuid1][1].output == "1234"
    assert responses[uuid1][1].stop_jobflow is True

    # check store has the activity output
    result1 = memory_jobstore.query_one({"uuid": uuid1})

    assert result1["output"] == "1234"


def test_stop_children_flow(memory_jobstore, clean_dir, stop_children_flow):
    from jobflow import run_locally

    flow = stop_children_flow()
    uuid1 = flow.jobs[0].uuid
    uuid2 = flow.jobs[1].uuid
    uuid3 = flow.jobs[2].uuid

    # run with log
    responses = run_locally(flow, store=memory_jobstore)

    # check responses has been filled
    assert len(responses) == 2
    assert len(responses[uuid1]) == 1
    assert uuid2 not in responses
    assert responses[uuid1][1].output == "1234"
    assert responses[uuid1][1].stop_children is True
    assert responses[uuid3][1].output == "12345_end"

    # check store has the activity output
    result1 = memory_jobstore.query_one({"uuid": uuid1})
    result2 = memory_jobstore.query_one({"uuid": uuid2})
    result3 = memory_jobstore.query_one({"uuid": uuid3})

    assert result1["output"] == "1234"
    assert result2 is None
    assert result3["output"] == "12345_end"


def test_error_flow(memory_jobstore, clean_dir, error_flow, capsys):
    from jobflow import run_locally

    flow = error_flow()

    # run with log
    responses = run_locally(flow, store=memory_jobstore)

    # check responses has been filled
    assert len(responses) == 0

    captured = capsys.readouterr()
    assert "error_func failed with exception" in captured.out

    with pytest.raises(RuntimeError):
        run_locally(flow, store=memory_jobstore, ensure_success=True)


def test_stored_data_flow(memory_jobstore, clean_dir, stored_data_flow, capsys):
    from jobflow import run_locally

    flow = stored_data_flow()

    responses = run_locally(flow, store=memory_jobstore)
    captured = capsys.readouterr()

    # check responses has been filled
    assert len(responses) == 1
    assert "Response.stored_data is not supported" in captured.out


def test_detour_stop_flow(memory_jobstore, clean_dir, detour_stop_flow):
    from jobflow import run_locally

    flow = detour_stop_flow()
    uuid1 = flow.jobs[0].uuid
    uuid3 = flow.jobs[1].uuid

    # run with log
    responses = run_locally(flow, store=memory_jobstore)
    uuid2 = [u for u in responses.keys() if u != uuid1 and u != uuid3][0]

    # check responses has been filled
    assert len(responses) == 2
    assert responses[uuid1][1].output == 11
    assert responses[uuid1][1].detour is not None
    assert responses[uuid2][1].output == "1234"

    # check store has the activity output
    result1 = memory_jobstore.query_one({"uuid": uuid1})
    result2 = memory_jobstore.query_one({"uuid": uuid2})
    result3 = memory_jobstore.query_one({"uuid": uuid3})

    assert result1["output"] == 11
    assert result2["output"] == "1234"
    assert result3 is None
